import pandas as pd

import pemi
import pemi.pd_mapper
import pemi.pipes.patterns

class SaSqlSourcePipe(pemi.pipes.patterns.SourcePipe):
    def __init__(self, *, sql, engine, **params):
        super().__init__(**params)

        self.sql = sql
        self.engine = engine
        self.field_maps = pemi.pd_mapper.schema_maps(self.schema)

    def extract(self):
        pemi.log.info("Extracting '%s' via:\n%s", self.name, self.sql)
        with self.engine.connect() as conn:
            sql_df = pd.read_sql(self.sql, conn)

        return sql_df

    def parse(self, data):
        pemi.log.info("Parsing '%s' results", self.name)
        mapper = pemi.pd_mapper.PdMapper(data, maps=self.field_maps).apply()
        self.targets['main'].df = mapper.mapped_df
        self.targets['errors'].df = mapper.errors_df
        return self.targets['main'].df

    def flow(self):
        self.parse(self.extract())



import pandas as pd

import pemi
import pemi.pipes.patterns
import pemi.transforms
from pemi.pd_mapper import *

class PdForkPipe(pemi.pipes.patterns.ForkPipe):
    def flow(self):
        for target in self.targets.values():
            target.df = self.sources['main'].df.copy()


class PdConcatPipe(pemi.pipes.patterns.ConcatPipe):
    def __init__(self, *, concat_opts={}, **params):
        super().__init__(**params)
        self.concat_opts = concat_opts


    def flow(self):
        source_dfs = [source.df for source in self.sources.values() if source.df is not None]
        if len(source_dfs) == 0:
            self.targets['main'].df = pd.DataFrame()
        else:
            self.targets['main'].df = pd.concat(source_dfs, **self.concat_opts)

# TODOC: Note that RowHandler('recode') will not work here
class PdLookupJoinPipe(pemi.Pipe):
    def __init__(self, main_key, lookup_key,
                 suffixes=('', '_lkp'),
                 missing_handler=None,
                 indicator=None,
                 lookup_prefix='',
                 fillna=None,
                 **kwargs): #pylint: disable=too-many-arguments
        super().__init__(**kwargs)

        self.main_key = main_key
        self.lookup_key = lookup_key
        self.suffixes = suffixes
        self.missing_handler = missing_handler or RowHandler('exclude')
        self.indicator = indicator
        self.lookup_prefix = lookup_prefix
        self.fillna = fillna

        self.source(
            pemi.PdDataSubject,
            name='main'
        )

        self.source(
            pemi.PdDataSubject,
            name='lookup'
        )

        self.target(
            pemi.PdDataSubject,
            name='main'
        )

        self.target(
            pemi.PdDataSubject,
            name='errors'
        )

    def flow(self):
        pemi.log.debug('PdLookupJoinPipe - main source columns: %s',
                       self.sources['main'].df.columns)
        pemi.log.debug('PdLookupJoinPipe - main lookup columns: %s',
                       self.sources['lookup'].df.columns)

        lkp_df = self.sources['lookup'].df

        if self.lookup_prefix != '':
            lkp_df = lkp_df.rename(
                columns={col: self.lookup_prefix + col
                         for col in lkp_df.columns if not col in self.lookup_key},
            )

        missing_keys = lkp_df[self.lookup_key].apply(
            lambda v: v.apply(pemi.transforms.isblank).any(), axis=1
        )
        if len(missing_keys) > 0:
            lkp_df = lkp_df[~missing_keys]

        uniq_lkp_df = lkp_df.sort_values(
            self.lookup_key
        ).groupby(
            self.lookup_key
        ).first().reset_index()

        merged_df = pd.merge(
            self.sources['main'].df,
            uniq_lkp_df,
            left_on=self.main_key,
            right_on=self.lookup_key,
            how='left',
            suffixes=self.suffixes,
            indicator='__indicator__'
        )

        def raise_on_mismatch(row):
            if row['__indicator__'] == 'left_only':
                raise KeyError('Lookup key "{}" not found'.format(dict(row[self.main_key])))
            return row


        mapper = PdMapper(merged_df, mapped_df=merged_df, maps=[
            PdMap(source=[*self.main_key, '__indicator__'], transform=raise_on_mismatch,
                  handler=self.missing_handler)
        ]).apply()

        if not self.indicator:
            del mapper.mapped_df['__indicator__']
        else:
            indicator_map = lambda v: True if v == 'both' else False
            mapper.mapped_df[self.indicator] = mapper.mapped_df['__indicator__'].apply(
                indicator_map
            ).astype('bool')

        if self.fillna:
            mapper.mapped_df.fillna(**self.fillna, inplace=True)

        self.targets['main'].df = mapper.mapped_df
        self.targets['errors'].df = mapper.errors_df

        pemi.log.debug('PdLookupJoinPipe - main target columns: %s',
                       self.targets['main'].df.columns)



class PdFieldValueForkPipe(pemi.Pipe):
    def __init__(self, field, forks, **kwargs):
        super().__init__(**kwargs)

        self.field = field
        self.forks = forks

        self.source(
            pemi.PdDataSubject,
            name='main'
        )

        for fork in self.forks:
            self.target(
                pemi.PdDataSubject,
                name=fork
            )

        self.target(
            pemi.PdDataSubject,
            name='remainder'
        )

    def flow(self):
        grouped = self.sources['main'].df.groupby(self.field)

        for fork in self.forks:
            if fork in grouped.groups:
                self.targets[fork].df = grouped.get_group(fork).copy()
            else:
                self.targets[fork].df = pd.DataFrame(columns=self.sources['main'].df.columns)

        remainder = set(grouped.groups.keys()) - set(self.forks)
        if len(remainder) > 0:
            self.targets['remainder'].df = pd.concat(
                [grouped.get_group(r) for r in remainder]
            ).sort_index()



import pickle
import copy
from collections import OrderedDict

import pemi
import pemi.connections


class Pipe():
    '''
    A pipe is a parameterized collection of sources and targets which can be executed (flow).
    '''

    def __init__(self, *, name='self', **params):
        self.name = name
        self.params = params
        self.sources = OrderedDict()
        self.targets = OrderedDict()
        self.pipes = OrderedDict()
        self.connections = pemi.connections.PipeConnections()

        # TODOC: special case of "self" pipe
        self.pipes['self'] = self

    def source(self, subject_class, name, schema=pemi.Schema(), **kwargs):
        self.sources[name] = subject_class(
            pipe=self,
            name=name,
            schema=schema,
            **kwargs
        )

    def target(self, subject_class, name, schema=pemi.Schema(), **kwargs):
        self.targets[name] = subject_class(
            pipe=self,
            name=name,
            schema=schema,
            **kwargs
        )

    def pipe(self, name, pipe):
        pipe.name = name
        self.pipes[name] = pipe


    def connect(self, from_pipe_name, from_subject_name):
        conn = pemi.connections.PipeConnection(self, from_pipe_name, from_subject_name)

        self.connections.append(conn)
        return conn


    def to_pickle(self, picklepipe=None):
        picklepipe = picklepipe or Pipe()

        for name, source in self.sources.items():
            psource = copy.copy(source)
            psource.pipe = picklepipe
            picklepipe.source(psource.__class__, name=name)
            picklepipe.sources[name] = psource

        for name, target in self.targets.items():
            ptarget = copy.copy(target)
            ptarget.pipe = picklepipe
            picklepipe.target(ptarget.__class__, name=name)
            picklepipe.targets[name] = ptarget

        for name, nestedpipe in self.pipes.items():
            if nestedpipe == self: continue

            nestedpicklepipe = Pipe()
            picklepipe.pipe(name=name, pipe=nestedpicklepipe)
            picklepipe.pipes[name] = nestedpipe.to_pickle(nestedpicklepipe)

        return pickle.dumps(picklepipe)



    def from_pickle(self, picklepipe=None):
        picklepipe = pickle.loads(picklepipe)

        for name, source in picklepipe.sources.items():
            source.pipe = self
            self.sources[name] = source

        for name, target in picklepipe.targets.items():
            target.pipe = self
            self.targets[name] = target

        for name, nestedpicklepipe in picklepipe.pipes.items():
            if nestedpicklepipe == picklepipe: continue

            self.pipes[name].from_pickle(nestedpicklepipe)

        return self


    def flow(self):
        raise NotImplementedError

    def __str__(self):
        return "<{}({}) {}>".format(self.__class__.__name__, self.name, id(self))



import pandas as pd

import pemi
import pemi.transforms

__all__ = [
    'RowHandler',
    'PdMapper',
    'PdMap'
]


class RowHandler:
    def __init__(self, mode='raise', recode=None):
        # ignore - Errors get returned as None and they don't go into the errors datadaset
        # warn - Errors get returned as None, but they still go into the errors datadaset
        # exclude - The entire record is excluded from the results, put into the errors dataset
        # recode - Alternate transform applied, records still go into the errors dataset
        # raise - Full stop, all subsequent processing halts

        self.mode = mode
        self.recode = recode
        self.errors = []

        self.handler = getattr(self, '_{}'.format(mode))

    def apply(self, transform, arg, idx):
        try:
            return transform(arg)
        except Exception as err: #pylint: disable=broad-except
            return self.handler(err, arg, idx)

    def mapping_error(self, err, idx):
        return {
            'mode': self.mode,
            'index': idx,
            'type': err.__class__.__name__,
            'message': str(err)
        }

    def catch_error(self, err, idx):
        self.errors.append(self.mapping_error(err, idx))

    def _raise(self, err, arg, idx): #pylint: disable=unused-argument
        pemi.log.error(self.mapping_error(err, idx))
        raise err

    def _recode(self, err, arg, idx):
        self.catch_error(err, idx)
        return self.recode(arg)

    def _warn(self, err, arg, idx): #pylint: disable=unused-argument
        self.catch_error(err, idx)
        pemi.log.warning(self.mapping_error(err, idx))
        return None

    def _ignore(self, err, arg, idx): #pylint: disable=unused-argument,no-self-use
        return None

    def _exclude(self, err, arg, idx): #pylint: disable=unused-argument
        self.catch_error(err, idx)
        pemi.log.error(self.mapping_error(err, idx))
        return None


class PdMapper:
    def __init__(self, source_df, mapped_df=None, maps=None):
        self.source_df = source_df
        if mapped_df is None:
            self.mapped_df = pd.DataFrame(index=self.source_df.index)
        else:
            self.mapped_df = mapped_df

        self.errors_df = pd.DataFrame([])
        self.errors = None
        self.maps = maps or []

    def apply(self):
        for pdmap in self.maps:
            pdmap(source_df=self.source_df, mapped_df=self.mapped_df)

        self.collect_errors()
        self.build_errors_df()
        self.exclude_errors()
        return self

    def collect_errors(self):
        self.errors = [err for pdmap in self.maps for err in pdmap.handler.errors]

    def build_errors_df(self):
        index = [err['index'] for err in self.errors]
        df = pd.DataFrame(self.errors, index=index)
        renamer = {k: '__error_{}__'.format(k) for k in df.columns}
        df.rename(columns=renamer, inplace=True)

        self.errors_df = df.join(self.source_df, how='left')

    def exclude_errors(self):
        exclude_idx = set([err['index'] for err in self.errors if err['mode'] == 'exclude'])
        if len(exclude_idx) > 0:
            self.mapped_df.drop(exclude_idx, inplace=True)

# TODO: I think pytlint is trying to tell me something here.....
class PdMap: #pylint: disable=too-many-instance-attributes,too-few-public-methods
    def __init__(self, source=None, target=None, transform=None, handler=None, **kwargs):
        if isinstance(source, str):
            self.source = source or []
            source = [source]

        if isinstance(target, str):
            self.target = target or []
            target = [target]

        self.sources = list(source or [])
        self.targets = list(target or [])
        self.transform = transform
        self.handler = handler or RowHandler()
        self.kwargs = kwargs

        self.source_df = None
        self.mapped_df = None
        self.error_df = None

        if len(self.sources) == 1 and len(self.targets) == 1 and self.transform is None:
            self._apply = getattr(self, '_apply_copy')
        elif len(self.sources) == 1 and len(self.targets) == 1:
            self._apply = getattr(self, '_apply_one_to_one')
        elif len(self.sources) == 1 and len(self.targets) == 0:
            self._apply = getattr(self, '_apply_one_to_zero')
        elif len(self.sources) == 0 and len(self.targets) == 1:
            self._apply = getattr(self, '_apply_zero_to_one')
        elif len(self.targets) == 1:
            self._apply = getattr(self, '_apply_many_to_one')
        else:
            self._apply = getattr(self, '_apply_many_to_many')

    def _transform(self, arg):
        return self.transform(arg)

    def _transform_one_to_one(self, row):
        return self.handler.apply(self._transform, row[self.source], row.name)

    def _transform_zero_to_one(self, row):
        return self.handler.apply(self._transform, row['__none__'], row.name)

    def _transform_many_to_one(self, row):
        return self.handler.apply(self._transform, row, row.name)

    def _transform_many_to_many(self, row):
        return self.handler.apply(self._transform, row, row.name)

    def apply(self):
        if len(self.source_df) > 0:
            self._apply()
        else:
            for target in self.targets:
                self.mapped_df[target] = None

    def _apply_copy(self):
        self.mapped_df[self.target] = self.source_df[self.source].copy()

    def _apply_one_to_one(self):
        self.mapped_df[self.target] = self.source_df[[self.source]].apply(
            self._transform_one_to_one, axis=1
        )

    def _apply_zero_to_one(self):
        empty_df = pd.DataFrame([], index=self.source_df.index)
        empty_df['__none__'] = None
        self.mapped_df[self.target] = empty_df.apply(self._transform_zero_to_one, axis=1)

    def _apply_one_to_zero(self):
        self.source_df[[self.source]].apply(self._transform_one_to_one, axis=1)

    def _apply_many_to_one(self):
        self.mapped_df[self.target] = self.source_df[self.sources].apply(
            self._transform_many_to_one, axis=1
        )

    def _apply_many_to_many(self):
        work_df = self.source_df[self.sources].apply(self._transform_many_to_many, axis=1)

        for target in self.targets:
            self.mapped_df[target] = work_df[target]

    def __call__(self, source_df, mapped_df=None, error_df=None):
        self.source_df = source_df
        self.mapped_df = mapped_df
        self.error_df = error_df
        self.apply()
        return self

def schema_maps(schema):
    field_maps = []
    for name, field in schema.items():
        field_map = PdMap(
            source=name, target=name,
            transform=field.coerce,
            handler=RowHandler('exclude')
        )
        field_maps.append(field_map)

        if field.metadata.get('allow_null', True) is False:
            field_map = PdMap(
                source=name, target=name,
                transform=pemi.transforms.validate_no_null(field),
                handler=RowHandler('exclude')
            )
            field_maps.append(field_map)
    return field_maps



import pandas as pd
import numpy as np
import pytest

import pemi
import pemi.data
import pemi.testing as pt
import pemi.pipes.pd
from pemi.fields import *
from pemi.pd_mapper import *

#TODO: Can I remove this?
#pylint: disable=redefined-outer-name

class PdLookupJoinPipeScenarioFactory():
    def __init__(self, scenario, pipe):
        self.scenario = scenario
        self.pipe = pipe

    @staticmethod
    def case_keys():
        ids = list(range(1000))
        for i in ids:
            yield {
                'main_source': {'key': 'k{}'.format(i)},
                'lookup': {'lkey': 'k{}'.format(i)},
                'main_target': {'key': 'k{}'.format(i)},
                'errors': {'key': 'k{}'.format(i)}
            }

    def scenario_setup(self):
        return {
            'runner': self.pipe.flow,
            'case_keys': self.case_keys(),
            'sources': {
                'main_source': pipe.sources['main'],
                'lookup': pipe.sources['lookup']
            },
            'targets': {
                'main_target': pipe.targets['main'],
                'errors': pipe.targets['errors']
            }
        }

    def ex_main(self):
        return pemi.data.Table(
            '''
            | key    | words |
            | -      | -     |
            | {k[1]} | words |
            | {k[1]} | words |
            | {k[3]} | more  |
            | {k[7]} | words |
            | {k[4]} | even  |
            | {k[4]} | more  |
            '''.format(k=self.scenario.case_keys.cache('main_source', 'key')),
            schema=pemi.Schema(
                key=StringField(),
                words=StringField()
            )
        )

    def ex_lookup(self):
        return pemi.data.Table(
            '''
            | lkey    | values | words  |
            | -       | -      | -      |
            | {k[1]}  | one    | I      |
            | {k[4]}  | four   | people |
            | {k[1]}  | ONE    | fnord  |
            | {k[3]}  | three  | dead   |
            | {k[4]}  | FOUR   | fnord  |
            | {k[2]}  | two    | see    |
            '''.format(k=self.scenario.case_keys.cache('lookup', 'lkey')),
            schema=pemi.Schema(
                lkey=StringField(),
                values=StringField(),
                words=StringField()
            )
        )

    def background(self):
        return [
            pt.when.example_for_source(self.scenario.sources['main_source'], self.ex_main()),
            pt.when.example_for_source(self.scenario.sources['lookup'], self.ex_lookup())
        ]



with pt.Scenario('PdLookupJoinPipe Basics') as scenario:
    pipe = pemi.pipes.pd.PdLookupJoinPipe(
        main_key=['key'],
        lookup_key=['lkey']
    )
    factory = PdLookupJoinPipeScenarioFactory(scenario, pipe)
    scenario.setup(**factory.scenario_setup())

    with scenario.case('it performs the lookup') as case:
        expected = pemi.data.Table(
            '''
            | key    | words | lkey     | values  | words_lkp |
            | -      | -     | -        | -       | -         |
            | {k[1]} | words | {lk[1]}  | one     | I         |
            | {k[1]} | words | {lk[1]}  | one     | I         |
            | {k[3]} | more  | {lk[3]}  | three   | dead      |
            | {k[4]} | even  | {lk[4]}  | four    | people    |
            | {k[4]} | more  | {lk[4]}  | four    | people    |
            '''.format(
                k=scenario.case_keys.cache('main_target', 'key'),
                lk=scenario.case_keys.cache('lookup', 'lkey')
            )
        )

        case.when(
            *factory.background()
        ).then(
            pt.then.target_matches_example(scenario.targets['main_target'], expected)
        )

    with scenario.case('it redirects errors') as case:
        expected = pemi.data.Table(
            '''
            | key    | words |
            | -      | -     |
            | {k[7]} | words |
            '''.format(k=scenario.case_keys.cache('errors', 'key'))
        )

        case.when(
            *factory.background()
        ).then(
            pt.then.target_matches_example(scenario.targets['errors'], expected)
        )


with pt.Scenario('PdLookupJoinPipe IgnoreHandler') as scenario:
    pipe = pemi.pipes.pd.PdLookupJoinPipe(
        main_key=['key'],
        lookup_key=['lkey'],
        missing_handler=RowHandler('ignore')
    )

    factory = PdLookupJoinPipeScenarioFactory(scenario, pipe)
    scenario.setup(**factory.scenario_setup())

    with scenario.case('it does not redirect errors') as case:
        case.when(
            *factory.background()
        ).then(
            pt.then.target_is_empty(scenario.targets['errors'])
        )

    with scenario.case('it works when the lookup is empty') as case:
        ex_lookup = pemi.data.Table(
            '''
            | lkey | values | words  |
            | -    | -      | -      |
            ''',
            schema=pemi.Schema(
                lkey=StringField(),
                values=StringField(),
                words=StringField()
            )
        )

        expected = pemi.data.Table(
            '''
            | key    | words |
            | -      | -     |
            | {k[1]} | words |
            | {k[1]} | words |
            | {k[3]} | more  |
            | {k[7]} | words |
            | {k[4]} | even  |
            | {k[4]} | more  |
            '''.format(
                k=scenario.case_keys.cache('main_target', 'key')
            )
        )

        case.when(
            *factory.background(),
            pt.when.example_for_source(scenario.sources['lookup'], ex_lookup)
        ).then(
            pt.then.target_is_empty(scenario.targets['errors'])
        )



with pt.Scenario('PdLookupJoinPipe FillNa') as scenario:
    pipe = pemi.pipes.pd.PdLookupJoinPipe(
        main_key=['key'],
        lookup_key=['lkey'],
        missing_handler=RowHandler('ignore'),
        fillna={'value': 'EMPTY'}
    )
    factory = PdLookupJoinPipeScenarioFactory(scenario, pipe)
    scenario.setup(**factory.scenario_setup())

    with scenario.case('it fills in missing values') as case:
        expected = pemi.data.Table(
            '''
            | key    | words | lkey    | values  | words_lkp |
            | -      | -     | -       | -       | -         |
            | {k[1]} | words | {lk[1]} | one     | I         |
            | {k[1]} | words | {lk[1]} | one     | I         |
            | {k[3]} | more  | {lk[3]} | three   | dead      |
            | {k[7]} | words | EMPTY   | EMPTY   | EMPTY     |
            | {k[4]} | even  | {lk[4]} | four    | people    |
            | {k[4]} | more  | {lk[4]} | four    | people    |
            '''.format(
                k=scenario.case_keys.cache('main_target', 'key'),
                lk=scenario.case_keys.cache('lookup', 'lkey')
            )
        )

        case.when(
            *factory.background()
        ).then(
            pt.then.target_matches_example(scenario.targets['main_target'], expected)
        )



with pt.Scenario('PdLookupJoinPipe Lookup Prefix') as scenario:
    pipe = pemi.pipes.pd.PdLookupJoinPipe(
        main_key=['key'],
        lookup_key=['lkey'],
        lookup_prefix='existing_'
    )

    factory = PdLookupJoinPipeScenarioFactory(scenario, pipe)
    scenario.setup(**factory.scenario_setup())

    with scenario.case('it prefixes lookup fields') as case:
        expected = pemi.data.Table(
            '''
            | key    | words | lkey    | existing_values | existing_words |
            | -      | -     | -       | -               | -              |
            | {k[1]} | words | {lk[1]} | one             | I              |
            | {k[1]} | words | {lk[1]} | one             | I              |
            | {k[3]} | more  | {lk[3]} | three           | dead           |
            | {k[4]} | even  | {lk[4]} | four            | people         |
            | {k[4]} | more  | {lk[4]} | four            | people         |
            '''.format(
                k=scenario.case_keys.cache('main_target', 'key'),
                lk=scenario.case_keys.cache('lookup', 'lkey')
            )
        )

        case.when(
            *factory.background()
        ).then(
            pt.then.target_matches_example(scenario.targets['main_target'], expected)
        )


with pt.Scenario('PdLookupJoinPipe Prefix Missing') as scenario:
    pipe = pemi.pipes.pd.PdLookupJoinPipe(
        main_key=['key'],
        lookup_key=['lkey'],
        lookup_prefix='existing_',
        missing_handler=RowHandler('ignore')
    )

    factory = PdLookupJoinPipeScenarioFactory(scenario, pipe)
    scenario.setup(**factory.scenario_setup())

    with scenario.case('it prefixes lookup fields when lookup is empty') as case:
        lookup = pemi.data.Table(
            '''
            | lkey | values | words  |
            | -    | -      | -      |
            ''',
            schema=pemi.Schema(
                lkey=StringField(),
                values=StringField(),
                words=StringField()
            )
        )

        expected = pemi.data.Table(
            '''
            | key    | words | lkey | existing_values | existing_words |
            | -      | -     | -    | -               | -              |
            | {k[1]} | words |      |                 |                |
            | {k[1]} | words |      |                 |                |
            | {k[3]} | more  |      |                 |                |
            | {k[7]} | words |      |                 |                |
            | {k[4]} | even  |      |                 |                |
            | {k[4]} | more  |      |                 |                |
            '''.format(
                k=scenario.case_keys.cache('main_target', 'key')
            ),
            schema=pemi.Schema(
                values=StringField(),
                words=StringField()
            )
        )

        case.when(
            *factory.background(),
            pt.when.example_for_source(scenario.sources['lookup'], lookup)
        ).then(
            pt.then.target_matches_example(scenario.targets['main_target'], expected)
        )



with pt.Scenario('PdLookupJoinPipe Indicator') as scenario:
    pipe = pemi.pipes.pd.PdLookupJoinPipe(
        main_key=['key'],
        lookup_key=['lkey'],
        missing_handler=RowHandler('ignore'),
        indicator='lkp_found'
    )

    factory = PdLookupJoinPipeScenarioFactory(scenario, pipe)
    scenario.setup(**factory.scenario_setup())

    with scenario.case('it adds an indicator') as case:
        expected = pemi.data.Table(
            '''
            | key    | words | lkp_found |
            | -      | -     | -         |
            | {k[1]} | words | True      |
            | {k[1]} | words | True      |
            | {k[3]} | more  | True      |
            | {k[7]} | words | False     |
            | {k[4]} | even  | True      |
            | {k[4]} | more  | True      |
            '''.format(
                k=scenario.case_keys.cache('main_target', 'key')
            )
        )

        case.when(
            *factory.background()
        ).then(
            pt.then.target_matches_example(scenario.targets['main_target'], expected)
        )



with pt.Scenario('PdLookupJoinPipe Blank Keys') as scenario:
    pipe = pemi.pipes.pd.PdLookupJoinPipe(
        main_key=['key'],
        lookup_key=['key']
    )

    def case_keys():
        ids = list(range(1000))
        for i in ids:
            yield {
                'main_source': {'alt_key': 'k{}'.format(i)},
                'main_target': {'alt_key': 'k{}'.format(i)},
                'errors': {'alt_key': 'k{}'.format(i)}
            }

    scenario.setup(
        runner=pipe.flow,
        case_keys=case_keys(),
        sources={
            'main_source': pipe.sources['main'],
            'lookup': pipe.sources['lookup']
        },
        targets={
            'main_target': pipe.targets['main'],
            'errors': pipe.targets['errors']
        }
    )

    with scenario.case('it does not use blanks as keys') as case:
        ex_main = pemi.data.Table(
            '''
            | key | alt_key |
            | -   | -       |
            | one | {k[1]}  |
            |     | {k[2]}  |
            | two | {k[3]}  |
            '''.format(k=scenario.case_keys.cache('main_source', 'alt_key')),
            schema=pemi.Schema(
                key=StringField(),
                alt_key=StringField()
            )
        )

        ex_lookup = pemi.data.Table(
            '''
            | key | value     |
            | -   | -         |
            | one | ONE       |
            | two | TWO       |
            |     | NOT THREE |
            |     | NOT FOUR  |
            ''',
            schema=pemi.Schema(
                key=StringField()
            )
        )

        expected = pemi.data.Table(
            '''
            | key | value | alt_key |
            | -   | -     | -       |
            | one | ONE   | {k[1]}  |
            | two | TWO   | {k[3]}  |
            '''.format(k=scenario.case_keys.cache('main_target', 'alt_key')),
            schema=pemi.Schema(
                key=StringField(),
                value=StringField(),
                alt_key=StringField()
            )
        )


        errors = pemi.data.Table(
            '''
            | key | alt_key |
            | -   | -       |
            |     | {k[2]}  |
            '''.format(k=scenario.case_keys.cache('main_target', 'alt_key')),
            schema=pemi.Schema(
                key=StringField(),
                alt_key=StringField()
            )
        )

        case.when(
            pt.when.example_for_source(scenario.sources['main_source'], ex_main),
            pt.when.example_for_source(scenario.sources['lookup'], ex_lookup)
        ).then(
            pt.then.target_matches_example(scenario.targets['main_target'], expected),
            pt.then.target_matches_example(scenario.targets['errors'], errors),
        )





class TestPdConcatPipe: #pylint: disable=no-self-use
    def test_it_concatenates_sources(self):
        pipe = pemi.pipes.pd.PdConcatPipe(sources=['s1', 's2'])
        pipe.sources['s1'].df = pd.DataFrame({
            'origin': ['s1', 's1', 's1'],
            'f1': [1, 2, 3]
        })

        pipe.sources['s2'].df = pd.DataFrame({
            'origin': ['s2', 's2'],
            'f2': [1, 2]
        })

        pipe.flow()
        expected_df = pd.DataFrame({
            'origin': ['s1', 's1', 's1', 's2', 's2'],
            'f1': [1, 2, 3, np.nan, np.nan],
            'f2': [np.nan, np.nan, np.nan, 1, 2]
        }, index=[0, 1, 2, 0, 1])
        actual_df = pipe.targets['main'].df
        pt.assert_frame_equal(actual_df, expected_df)

    def test_given_no_data(self):
        'It returns an empty dataframe'
        pipe = pemi.pipes.pd.PdConcatPipe(sources=['s1', 's2'])
        pipe.flow()

        expected_df = pd.DataFrame()
        actual_df = pipe.targets['main'].df
        pt.assert_frame_equal(actual_df, expected_df)


class TestPdFieldValueForkPipe: #pylint: disable=no-self-use
    @pytest.fixture(scope='class')
    def pipe(self):
        pipe = pemi.pipes.pd.PdFieldValueForkPipe(
            field='target',
            forks=['create', 'update', 'empty']
        )

        df = pd.DataFrame({
            'target': ['create', 'update', 'update', 'else1', 'create', 'else2'],
            'values': [1, 2, 3, 4, 5, 6]
        })

        pipe.sources['main'].df = df
        pipe.flow()
        return pipe

    def test_it_forks_data_to_create(self, pipe):
        expected_df = pd.DataFrame({
            'target': ['create', 'create'],
            'values': [1, 5]
        }, index=[0, 4])
        actual_df = pipe.targets['create'].df
        pt.assert_frame_equal(actual_df, expected_df)

    def test_it_forks_data_to_update(self, pipe):
        expected_df = pd.DataFrame({
            'target': ['update', 'update'],
            'values': [2, 3]
        }, index=[1, 2])
        actual_df = pipe.targets['update'].df
        pt.assert_frame_equal(actual_df, expected_df)

    def test_it_puts_unknown_values_in_remainder(self, pipe):
        expected_df = pd.DataFrame({
            'target': ['else1', 'else2'],
            'values': [4, 6]
        }, index=[3, 5])
        actual_df = pipe.targets['remainder'].df
        pt.assert_frame_equal(actual_df, expected_df)

    def test_it_creates_an_empty_dataframe_with_the_right_columns(self, pipe):
        expected_df = pd.DataFrame(columns=['target', 'values'])
        actual_df = pipe.targets['empty'].df
        pt.assert_frame_equal(actual_df, expected_df)



import pandas as pd

# TODO: prefix, postfix, ifblank, blankif

def validate_no_null(field):
    def _validate(value):
        if value == field.null:
            raise ValueError("null is not allowed for field '{}'".format(field.name))
        return value
    return _validate

def isblank(value):
    return value is not False and (value is None or pd.isnull(value) or not value)

def concatenate(delimiter=''):
    def _concatenate(row):
        return delimiter.join(row)
    return _concatenate

def nvl(default=''):
    def _nvl(row):
        return next((v for v in row if not isblank(v)), default)
    return _nvl



import re
import io
import random

import pandas as pd
from faker import Factory

import pemi
from pemi.fields import *


fake = Factory.create() #pylint: disable=invalid-name


class UniqueIdGenerator: #pylint: disable=too-few-public-methods
    '''
    Class used to build id generators.
    fmt - A function that accepts a single integer argument and returns a value to be used as an id.

    Example:
      students = UniqueIdGenerator(lambda i: 'S{i}'.format(i))
      [next(students) for x in range(5)]
      #=> ['S6', 'S5', 'S3', 'S1', 'S4', 'S7', 'S9', 'S2', 'S8', 'S16']
    '''
    def __init__(self, fmt):
        self.fmt = fmt
        self.size = 1
        self.gen_sample()

    def gen_sample(self):
        self.sample = list(range(10**(self.size-1), 10**self.size))
        random.shuffle(self.sample)
        self.size += 1

    def __next__(self):
        i = self.sample.pop()
        if len(self.sample) == 0:
            self.gen_sample()
        return self.fmt(i)


class InvalidHeaderSeparatorError(Exception): pass

default_fakers = { #pylint: disable=invalid-name
    IntegerField:  fake.pyint, #pylint: disable=no-member
    StringField:   fake.word, #pylint: disable=no-member
    DateField:     fake.date_object, #pylint: disable=no-member
    DateTimeField: fake.date_time, #pylint: disable=no-member
    FloatField:    fake.pyfloat, #pylint: disable=no-member
    DecimalField:  fake.pydecimal, #pylint: disable=no-member
    BooleanField:  fake.pybool, #pylint: disable=no-member
    JsonField:     lambda: fake.pydict(5, True, 'str', 'int', 'date') #pylint: disable=no-member
}


class Table: #pylint: disable=too-many-arguments,too-few-public-methods
    def __init__(self, markdown=None, nrows=10,
                 schema=pemi.Schema(), fake_with=None, coerce_with=None):
        self.markdown = markdown
        self.schema = schema
        self.nrows = nrows
        self.fake_with = fake_with or {}
        self.coerce_with = coerce_with or {}

        if self.markdown:
            self.defined_fields = list(self._build_from_markdown().columns)
        else:
            self.defined_fields = list(self.schema.keys())

    def _clean_markdown(self):
        cleaned = self.markdown

        # Remove trailing comments
        cleaned = re.compile(r'(#.*$)', flags=re.MULTILINE).sub('', cleaned)

        # Remove whitespace surrouding pipes
        cleaned = re.compile(r'[ \t]*\|[ \t]*').sub('|', cleaned)

        # Remove beginning and terminal pipe on each row
        cleaned = re.compile(r'(^\s*\|\s*|\s*\|\s*$)', flags=re.MULTILINE).sub('', cleaned)

        # Split by newlines
        cleaned = cleaned.split('\n')

        # Remove header separator
        header_separator = cleaned.pop(1)
        if re.search(re.compile(r'^[\s\-\|]*$'), header_separator) is None:
            raise InvalidHeaderSeparatorError('Bad header separator: {}'.format(header_separator))

        # Unsplit
        cleaned = '\n'.join(cleaned)
        return cleaned

    def _build_from_markdown(self):
        cleaned = self._clean_markdown()
        str_df = pd.read_csv(
            io.StringIO(cleaned),
            sep='|',
            converters={k:str for k in self.schema.keys()}
        )

        df = pd.DataFrame()
        for header in list(str_df):
            if header in self.coerce_with:
                df[header] = str_df[header].apply(self.coerce_with[header])
            elif header in self.schema.keys():
                df[header] = str_df[header].apply(self.schema[header].coerce)
            else:
                df[header] = str_df[header]
        return df

    def _fake_series(self, column, nsample=5):
        default_faker = default_fakers.get(type(self.schema[column]))
        valid_faker = self.fake_with.get(column, {}).get('valid')
        faker_func = valid_faker or default_faker

        if self.fake_with.get(column, {}).get('unique'):
            fake_data_dupes = [faker_func() for i in range(nsample * 3)]
            fake_data = list(set(fake_data_dupes))[0:nsample]
            random.shuffle(fake_data)
        else:
            fake_data = [faker_func() for i in range(nsample)]

        return pd.Series(fake_data)


    @property
    def df(self):
        if self.markdown:
            df = self._build_from_markdown()
        else:
            df = pd.DataFrame([], index=range(self.nrows))

        for column in self.schema.keys():
            if column not in df:
                df[column] = self._fake_series(column, len(df))
        return df



import os

import pytest

import sqlalchemy as sa

import pemi
import pemi.testing as pt
import pemi.pipes.sa
from pemi.fields import *

sa_engine = sa.create_engine('postgresql://{user}:{password}@{host}/{dbname}'.format(
    user=os.environ.get('POSTGRES_USER'),
    password=os.environ.get('POSTGRES_PASSWORD'),
    host=os.environ.get('POSTGRES_HOST'),
    dbname=os.environ.get('POSTGRES_DB')
))


def truncate_sa_sources(scenario_sources):
    sa_sources = [
        source.subject
        for source in scenario_sources if isinstance(source.subject, pemi.SaDataSubject)
    ]
    for source in sa_sources:
        with source.engine.connect().begin() as trans:
            trans.connection.execute('TRUNCATE {}'.format(source.table))

#pylint: disable=unused-argument
@pytest.fixture(scope='module')
def db_schema_init():
    with sa_engine.connect() as conn:
        conn.execute(
            '''
            DROP TABLE IF EXISTS sales_fact;
            CREATE TABLE sales_fact (
              beer_id INT,
              name VARCHAR(80),
              sold_at DATE,
              quantity INT
            );
            '''
        )


@pytest.fixture
def db_case_clean(case, db_schema_init):
    if not case.scenario.has_run:
        truncate_sa_sources(case.scenario.sources.values())
        yield
        truncate_sa_sources(case.scenario.sources.values())
    else:
        yield
#pylint: enable=unused-argument

with pt.Scenario('SaSqlSourcePipe', usefixtures=['db_case_clean']) as scenario:
    sales_schema = pemi.Schema(
        beer_id=IntegerField(),
        name=StringField(),
        sold_at=DateField(),
        quantity=IntegerField()
    )

    pipe = pemi.Pipe()
    pipe.pipe(
        name='sql_source',
        pipe=pemi.pipes.sa.SaSqlSourcePipe(
            engine=sa_engine,
            schema=sales_schema,
            sql='SELECT * FROM sales_fact;'
        )
    )

    pipe.source(
        pemi.SaDataSubject,
        name='sales',
        schema=sales_schema,
        engine=sa_engine,
        table='sales_fact'
    )


    def case_keys():
        ids = list(range(1000))
        for i in ids:
            yield {
                'sales': {'beer_id': i},
                'sql_source': {'beer_id': i}
            }

    scenario.setup(
        runner=pipe.pipes['sql_source'].flow,
        case_keys=case_keys(),
        sources={
            'sales': pipe.sources['sales']
        },
        targets={
            'sql_source': pipe.pipes['sql_source'].targets['main']
        }
    )

    with scenario.case('it queries data stored') as case:
        ex_sales = pemi.data.Table(
            '''
            | beer_id | sold_at    | quantity |
            | -       | -          | -        |
            | {b[1]}  | 2017-01-01 | 3        |
            | {b[2]}  | 2017-01-02 | 2        |
            | {b[3]}  | 2017-01-03 | 5        |
            '''.format(b=scenario.case_keys.cache('sales', 'beer_id')),
            schema=pipe.sources['sales'].schema
        )

        case.when(
            pt.when.example_for_source(scenario.sources['sales'], ex_sales)
        ).then(
            pt.then.target_matches_example(scenario.targets['sql_source'], ex_sales)
        )

    with scenario.case('it works with na') as case:
        ex_sales = pemi.data.Table(
            '''
            | beer_id | sold_at    | quantity |
            | -       | -          | -        |
            | {b[1]}  | 2017-01-01 | 3        |
            | {b[2]}  | 2017-01-02 |          |
            | {b[3]}  | 2017-01-03 | 5        |
            '''.format(b=scenario.case_keys.cache('sales', 'beer_id')),
            schema=pipe.sources['sales'].schema
        )

        case.when(
            pt.when.example_for_source(scenario.sources['sales'], ex_sales)
        ).then(
            pt.then.target_matches_example(scenario.targets['sql_source'], ex_sales)
        )



import copy

from collections import OrderedDict

class Schema:
    def __init__(self, *args, **kwargs):
        self.fields = OrderedDict()
        for field in args:
            self.fields[field.name] = field

        for name, field in kwargs.items():
            field.name = name
            self.fields[name] = field

    def __getitem__(self, key):
        if isinstance(key, list):
            return Schema(**{self[k].name: self[k] for k in key})
        return self.fields[key]

    def keys(self):
        return list(self.fields.keys())

    def values(self):
        return list(self.fields.values())

    def items(self):
        return self.fields.items()

    def coercions(self):
        return {f.name: f.coerce for f in self.fields.values()}

    def __str__(self):
        return "\n".join(
            ['{} -> {}'.format(name, meta.__str__()) for name, meta in self.fields.items()]
        )

    def __repr__(self):
        return '<{}({}):\n{}\n>'.format(self.__class__.__name__, id(self), self.__str__())

    def __eq__(self, other):
        return self.fields == other.fields

    def merge(self, other):
        merged_fields = {**self.fields, **other.fields}

        for name, field in merged_fields.items():
            if name in self.fields:
                self_meta = copy.deepcopy(self.fields[name].metadata)
            else:
                self_meta = {}

            if name in other.fields:
                other_meta = copy.deepcopy(other.fields[name].metadata)
            else:
                other_meta = {}

            merged_field_metadata = {**self_meta, **other_meta}
            merged_fields[name] = type(field)(**merged_field_metadata)

        return Schema(**merged_fields)

    def copy(self):
        return copy.deepcopy(self)

    def rename(self, mapper):
        new_fields = []
        for field in self.fields.values():
            new_field = copy.deepcopy(field)
            new_field.name = mapper.get(field.name, field.name)
            new_fields.append(new_field)
        return Schema(*new_fields)

    def select(self, func):
        'Returns a new schema with the fields selected via a function (func) of the field'
        return Schema(**{name:copy.deepcopy(field) for name, field in self.items() if func(field)})
